use clap::{Parser, ValueEnum};
use psxcdgen_ex::{encoder::{EncodingFunctions, psx::{calculate_psx_lbas, calculate_psx_length_for, encode_psx_image}}, file_writer::{ImageType, write_image}, config_reader};
use std::{path::PathBuf, };
use tool_helper::Error;

#[derive(Parser)]
#[clap(about = "Creates an ISO image from a description file", long_about = None)]
struct CommandLine {
    #[clap(value_enum, value_parser, help="Specifies the system for which to create the disc image")]
    system_type: SystemType,

    #[clap(value_enum, value_parser, help="Specifies the disc image type")]
    output_type: ImageType,

    #[clap(short='o', help="Output path; Some OUTPUT_TYPE might create additional files with different endings")]
    output_file: PathBuf,

    #[clap(long="list", id="Path to file", help="If set will list the CD content to stdout; With path will output list to file")]
    list_content: Option<Option<PathBuf>>,

    #[clap(value_parser, help="Input XML path for creating the image")]
    input_file: PathBuf,
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, ValueEnum)]
enum SystemType {
    Psx
}

impl SystemType {
    pub fn get_encoding_functions(self) -> EncodingFunctions {
        match self {
            SystemType::Psx => EncodingFunctions{length_calculator: calculate_psx_length_for, lba_calculator: calculate_psx_lbas, encoder: encode_psx_image}
        }
    }
}

fn run_main(cmd_line: CommandLine) -> Result<(), Error> {
    let encoding_functions         = cmd_line.system_type.get_encoding_functions();
    let (desc, lba_embedded_files) = psxcdgen_ex::process(config_reader::parse_xml(std::fs::read_to_string(cmd_line.input_file)?)?, encoding_functions.lba_calculator)?;
    let file_map                   = desc.create_file_map();

    psxcdgen_ex::process_files(file_map, lba_embedded_files, encoding_functions.length_calculator)?;
    write_image(&desc, encoding_functions.encoder, cmd_line.output_type, cmd_line.output_file)?;

    if let Some(list_content_option) = cmd_line.list_content {
        psxcdgen_ex::dump_content(&desc, tool_helper::open_output(list_content_option)?)
    }
    
    else {
        Ok(())
    }
}

fn main() {
    match CommandLine::try_parse() {
        Ok(cmd_line) => {
            match run_main(cmd_line) {
                Ok(_)      => (),
                Err(error) => eprintln!("{}", error)
            }
        },
        Err(error) => {
            eprintln!("{}", error);
        }
    }
}